classdef eluLayer < nnet.layer.Layer
% elu layer with learnable parameter. copied from the forum example. 
% ys

    properties (Learnable)
        alpha
    end
    
    methods
        function layer = eluLayer(num_channels,name)
            layer.Type = 'Exponential Linear Unit';
            
            % Assign layer name if it is passed in.
            if nargin > 1
                layer.Name = name;
            end
            
            % Give the layer a meaningful description.
            layer.Description = "Exponential linear unit with " + ...
                num_channels + " channels";
            
            % Initialize the learnable alpha parameter.
            layer.alpha = rand(1,1,num_channels);
        end

        
       function Z = predict(layer, X)
            % Forward input data through the layer and output the result
            Z = max(0, X) + layer.alpha .*(exp(min(0, X))-1);
            
        end
        
        
        function [Z, memory] = forward(layer, X)
            % (Optional) Forward input data through the layer at training
            % time and output the result and a memory value
            %
            % Inputs:
            %         layer  - Layer to forward propagate through
            %         X      - Input data
            % Output:
            %         Z      - Output of layer forward function
            %         memory - Memory value which can be used for
            %                  backward propagation

            % Layer forward function for training goes here
            memory = exp(min(0, X))-1;
            Z = max(0, X) + layer.alpha .* memory;

        end
        
        
        function [dLdX, dLdalpha] = backward(layer, X, Z, dLdZ, memory)
            % Backward propagate the derivative of the loss function through 
            % the layer 
            % y = dLdZ
            
            dLdX = (layer.alpha+Z) .* dLdZ; % negative part => (a+f(x))*y
            dLdX(X>0) = dLdZ(X>0); % positive part => y
            
            % derivat the the 
            dLdalpha = memory .* dLdZ; % negative part only => exp(x)-1 
            dLdalpha = sum(sum(dLdalpha,1),2);
            
            % Sum over all observations in mini-batch
            dLdalpha = sum(dLdalpha,4);            
        end
    end
end